-- keybinding constants
BINDING_HEADER_TOOLTIPNOTES               = "Tooltip Notes"
BINDING_NAME_TOOLTIPNOTES_SET_TARGET      = "Set selected target's note"
BINDING_NAME_TOOLTIPNOTES_SHOW_TARGET     = "Show selected target's note"
BINDING_NAME_TOOLTIPNOTES_SAY_TARGET      = "Say selected target's note"
BINDING_NAME_TOOLTIPNOTES_CLEAR_TARGET    = "Clear selected target's info"
BINDING_NAME_TOOLTIPNOTES_SET_MOUSEOVER   = "Set mouseover target's note"
BINDING_NAME_TOOLTIPNOTES_SHOW_MOUSEOVER  = "Show mouseover target's note"
BINDING_NAME_TOOLTIPNOTES_SAY_MOUSEOVER   = "Say mouseover target's note"
BINDING_NAME_TOOLTIPNOTES_CLEAR_MOUSEOVER = "Clear mouseover target's info"

local string_format = string.format

----------------------------------------------------------------------------------------------------
-- Notes popup dialog
----------------------------------------------------------------------------------------------------
--------------------------------------------------
-- add a note (or clear it if blank)
--------------------------------------------------
local function AddNote(name, note)
	if not name then return end

	local properName = name:find("%u") and name
	name = name:lower()

	local noteList = TooltipNotesSave.creatures
	local creature = noteList[name]

	-- creature doesn't exist yet so add it
	if not creature then
		if note and note ~= "" then
			noteList[name] = {}
			noteList[name].customNote = note
			noteList[name].properName = properName
		end
		return
	end

	-- creature already exists - add/change or remove the note
	noteList[name].properName = noteList[name].properName or properName
	if note and note ~= "" then
		creature.customNote = note
	else
		-- removing the note - also remove the whole creature if no spells are recorded for them
		creature.customNote = nil
		if next(creature) == nil then
			noteList[name] = nil
		end
	end
end

--------------------------------------------------
-- open popup window to edit a note
--------------------------------------------------
local skipKey -- when opening from a binding, the first key needs to be erased
local function OpenNoteDialog(lowerName, properName, fromBinding)
	if not lowerName or lowerName == "" then return end

	-- create the dialog if it hasn't been yet
	if not StaticPopupDialogs["TOOLTIPNOTES_POPUP"] then
		StaticPopupDialogs["TOOLTIPNOTES_POPUP"] = {
			text = "",
			button1 = ACCEPT,
			button2 = CANCEL,
			timeout = 0,
			whileDead = 1,
			hideOnEscape = 1,
			hasEditBox = 1,
			editBoxWidth = 350,
			preferredIndex = 3, -- to avoid interferring with Blizzard UI popups
			OnShow = function(self)

				local creature = TooltipNotesSave.creatures[StaticPopupDialogs["TOOLTIPNOTES_POPUP"].text:lower()]
				_G[self:GetName().."EditBox"]:SetText(creature and creature.customNote or "")
				_G[self:GetName().."EditBox"]:SetFocus()
			end,
			-- Reset the text when opening from a key binding or else that key will be put in the editbox.
			-- There's probably a better way to stop that from happening.
			EditBoxOnTextChanged = function(self)
				if skipKey then
					skipKey = nil
					StaticPopupDialogs["TOOLTIPNOTES_POPUP"].OnShow(self:GetParent())
				end
			end,
			EditBoxOnEscapePressed = function(self)
				self:GetParent():Hide()
			end,
			OnAccept = function(self)
				local name = StaticPopupDialogs["TOOLTIPNOTES_POPUP"].text
				local note = _G[self:GetName().."EditBox"]:GetText()
				AddNote(name, note)
			end,
			EditBoxOnEnterPressed = function(self)
				StaticPopupDialogs["TOOLTIPNOTES_POPUP"].OnAccept(self:GetParent())
				self:GetParent():Hide()
			end,
		}
	end -- end of creating static popup

	skipKey = fromBinding
	local creature = TooltipNotesSave.creatures[lowerName]
	StaticPopupDialogs["TOOLTIPNOTES_POPUP"].text = creature and creature.properName or properName or lowerName
	StaticPopup_Show("TOOLTIPNOTES_POPUP")
end

----------------------------------------------------------------------------------------------------
-- /tnote command
----------------------------------------------------------------------------------------------------
--------------------------------------------------
-- find a target either by input text or targeting
--------------------------------------------------
local function GetSlashTarget(input)
	local target, properName
	if input == "mouseover" then
		target = UnitName("mouseover")
		properName = target
	elseif input and input ~= "" then
		target = input
	else
		target = UnitName("target")
		properName = target
	end
	if not target then
		if input ~= "mouseover" then
			DEFAULT_CHAT_FRAME:AddMessage("You must have a target selected or use /tnote set <name>.")
		end
		return
	end
	return target:lower(), properName
end

--------------------------------------------------
-- changing tooltip line colors
--------------------------------------------------
local colorPickerColor = nil -- color table for the picker - reference to one of the saved settings

-- called when the color changes or the picker is canceled
local function ColorPickerCallback(restore)
	if restore then
		colorPickerColor[1], colorPickerColor[2], colorPickerColor[3] = unpack(restore)
	else
		colorPickerColor[1], colorPickerColor[2], colorPickerColor[3] = ColorPickerFrame:GetColorRGB()
	end
end

-- open and show the color picker
local function ShowColorPicker(r, g, b)
	ColorPickerFrame:SetColorRGB(r,g,b)
	ColorPickerFrame.hasOpacity, ColorPickerFrame.opacity = nil, nil
	ColorPickerFrame.previousValues = {r,g,b}
	ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc =
		ColorPickerCallback, ColorPickerCallback, ColorPickerCallback
	ColorPickerFrame:Hide() -- in case it's already open, this will reopen it to update the color
	ColorPickerFrame:Show()
end

--------------------------------------------------
-- say or print the notes of a creature
--------------------------------------------------
local function SayOrPrintNotes(name, channel)
	local creature = TooltipNotesSave.creatures[name:lower()]
	if not creature then
		DEFAULT_CHAT_FRAME:AddMessage("There isn't any information about '" .. name .. "' to show.")
		return
	end

	name = creature.properName or name

	-- show custom note first
	if creature.customNote then
		if channel then
			SendChatMessage("Notes about " .. name .. ": " .. creature.customNote, channel)
		else
			DEFAULT_CHAT_FRAME:AddMessage(name .. ": " .. creature.customNote)
		end
	end

	-- spell list
	local _, _, difficulty = GetInstanceInfo()
	local spellInfo = creature[difficulty]
	if spellInfo then
		local links = {}
		for id in pairs(spellInfo) do
			if id ~= "tooltipString" then
				links[#links+1] = string_format("|cff71d5ff|Hspell:%d|h[%s]|h|r ", id, (GetSpellInfo(id)) or "Unknown")
			end
		end

		-- build a list of messages under 255 characters long each (unless printing)
		local messageList = { (creature.customNote and "Uses: " or (name .. " uses: ")) }
		local listCount = #messageList
		local messageLength = #messageList[listCount]
		local linkLength
		for _,link in ipairs(links) do
			linkLength = #link
			if channel and messageLength + linkLength > 255 then
				messageList[listCount+1] = link
				listCount = listCount + 1
				messageLength = linkLength
			else
				messageList[listCount] = messageList[listCount] .. link
				messageLength = messageLength + linkLength
			end
		end

		-- show or send the message list
		for i=1,listCount do
			if not channel then
				DEFAULT_CHAT_FRAME:AddMessage(messageList[i])
			else
				SendChatMessage(messageList[i], channel)
			end
		end
	end
end

--------------------------------------------------
-- /tnote command
--------------------------------------------------
_G.SLASH_TNOTE1 = "/tnote"
function SlashCmdList.TNOTE(input, fromBinding)
	input = input or ""

	local command, value = input:match("(%w+)%s*(.*)")
	if command then command = command:lower() end

	local noteSettings = TooltipNotesSave.settings

	-- /tnote set - set a note on something
	if command == "set" then
		local target, properName = GetSlashTarget(value)
		if target then
			OpenNoteDialog(target, properName, fromBinding)
		end
		return
	end

	-- /tnote mouseover - set a note on the mouseover target
	if command == "setmouseover" then
		local target, properName = GetSlashTarget("mouseover")
		if target then
			OpenNoteDialog(target, properName, fromBinding)
		end
		return
	end

	-- /tnote clear - remove all notes from something
	if command == "clear" or command == "clearmouseover" then
		local target, properName = GetSlashTarget(command == "clearmouseover" and "mouseover" or value)
		if target then
			local creature = TooltipNotesSave.creatures[target]
			if creature then
				TooltipNotesSave.creatures[target] = nil
				DEFAULT_CHAT_FRAME:AddMessage("Information for '" .. (properName or target) .. "' has been deleted.")
			else
				DEFAULT_CHAT_FRAME:AddMessage("There is nothing to clear for that target.")
			end
		end
		return
	end

	-- /tnote say - say note/spells to the raid/party
	if command == "say" or command == "saymouseover" then
		local target = GetSlashTarget(command == "saymouseover" and "mouseover" or value)
		if target then
			local channel = (IsInRaid() and "raid") or (IsInGroup() and "party")
			if channel then
				SayOrPrintNotes(target, channel)
			else
				DEFAULT_CHAT_FRAME:AddMessage("You aren't in a group.")
			end
		end
		return
	end

	-- /tnote show - print note/immunities to yourself
	if command == "show" or command == "showmouseover" then
		local target = GetSlashTarget(command == "showmouseover" and "mouseover" or value)
		if target then
			SayOrPrintNotes(target, nil)
		end
		return
	end

	-- /tnote notecolor
	if command == "notecolor" then
		colorPickerColor = noteSettings["noteColor"]
		ShowColorPicker(unpack(colorPickerColor))
		return
	end

	-- /tnote spellcolor
	if command == "spellcolor" then
		colorPickerColor = noteSettings["spellColor"]
		ShowColorPicker(unpack(colorPickerColor))
		return
	end

	-- /tnote wordwrap - set how long tooltip lines can be in notes and immunity lists
	if command == "wordwrap" then
		local value = value and tonumber(value) or nil
		if value and value < 0 then
			DEFAULT_CHAT_FRAME:AddMessage("The width number must be 0 (for no wrapping) or greater.")
		elseif value and value >= 0 then
			noteSettings.wordWrapWidth = value
			if noteSettings.wordWrapWidth == 0 then
				DEFAULT_CHAT_FRAME:AddMessage("Custom word wrapping has been disabled.")
			else
				DEFAULT_CHAT_FRAME:AddMessage("Custom word wrapping has been set to " ..
					noteSettings.wordWrapWidth .. " characters.")
			end
		else
			DEFAULT_CHAT_FRAME:AddMessage("Syntax: /tnote wordwrap <width or 0>")
			if noteSettings.wordWrapWidth > 0 then
				DEFAULT_CHAT_FRAME:AddMessage("Custom word wrapping is currently " ..
					noteSettings.wordWrapWidth .. " characters.")
			end
		end
		return
	end

	-- no command used - show syntax and some current settings
	DEFAULT_CHAT_FRAME:AddMessage('TooltipNotes commands:', 1, 1, 0)
	DEFAULT_CHAT_FRAME:AddMessage('/tnote set [name]')
	DEFAULT_CHAT_FRAME:AddMessage('/tnote clear [name]')
	DEFAULT_CHAT_FRAME:AddMessage('/tnote show [name]')
	DEFAULT_CHAT_FRAME:AddMessage('/tnote say [name]')
	DEFAULT_CHAT_FRAME:AddMessage('/tnote notecolor')
	DEFAULT_CHAT_FRAME:AddMessage('/tnote spellcolor')
	DEFAULT_CHAT_FRAME:AddMessage('/tnote wordwrap <width>')
	if noteSettings.wordWrapWidth > 0 then
		DEFAULT_CHAT_FRAME:AddMessage(" ")
		DEFAULT_CHAT_FRAME:AddMessage("Custom word wrapping has been set to " ..
			noteSettings.wordWrapWidth .. " characters.")
	end
end
