-- saved settings table - defaults set up in TooltipNotes_OnEventLoad
-- TooltipNotesSave.settings[setting] = value
-- TooltipNotesSave.creatures["name"].properName = "Proper Capitalized Name"
-- TooltipNotesSave.creatures["name"].customNote = "custom note"
-- TooltipNotesSave.creatures["name"][difficulty].tooltipString = "Spell, Spell 2"
-- TooltipNotesSave.creatures["name"][difficulty][spellId] = true
TooltipNotesSave   = nil
local noteSettings = nil -- reference to TooltipNotesSave.settings
local noteList     = nil -- reference to TooltipNotesSave.creatures

-- frame used for listening to events
local eventFrame = CreateFrame("frame")
eventFrame:Hide()

-- local references
local table_insert = table.insert
local table_concat = table.concat

----------------------------------------------------------------------------------------------------
-- Helper functions
----------------------------------------------------------------------------------------------------
-- word wrap for tooltip lines - from http://rosettacode.org/wiki/Word_wrap#Lua
local function SplitTokens(text)
	local res = {}
	for word in text:gmatch("%S+") do
		res[#res+1] = word
	end
	return res
end

function WordWrap(text, linewidth)
	if linewidth == 0 then
		return {text}
	end
	if not linewidth then
		linewidth = 60
	end

	local spaceleft = linewidth
	local res = {}
	local line = {}
	for _, word in ipairs(SplitTokens(text)) do
		if #word + 1 > spaceleft then
			table_insert(res, table_concat(line, ' '))
			line = {word}
			spaceleft = linewidth - #word
		else
			table_insert(line, word)
			spaceleft = spaceleft - (#word + 1)
		end
	end
	table_insert(res, table_concat(line, ' '))
	return res
end

----------------------------------------------------------------------------------------------------
-- watch combat events to find spells
----------------------------------------------------------------------------------------------------
-- table of combat log actions to look for
local spellActions = {
	["SPELL_DAMAGE"]=1,
	["SPELL_MISSED"]=1,
	["SPELL_HEAL"]=1,
	["SPELL_ENERGIZE"]=1,
	["SPELL_DRAIN"]=1,
	["SPELL_LEECH"]=1,
	["SPELL_INTERRUPT"]=1,
	["SPELL_EXTRA_ATTACKS"]=1,
	["SPELL_AURA_APPLIED"]=1,
	["SPELL_AURA_APPLIED_DOSE"]=1,
	["SPELL_AURA_REFRESH"]=1,
	["SPELL_CAST_START"]=1,
	["SPELL_CAST_SUCCESS"]=1,
	["SPELL_CAST_FAILED"]=1,
}

function TooltipNotes_OnEventCombat(self, event)
	local _, action, _, casterGUID, casterName, _, _, _, _, _, _, spellId, spellName = CombatLogGetCurrentEventInfo()
	if not casterGUID or not spellActions[action] or not casterGUID:find("^Creature%-") or not spellName or spellName == "Dazed" then
		return
	end

	local creature = noteList[casterName:lower()]
	if not creature then
		local lowerName = casterName:lower()
		noteList[lowerName] = {}
		creature = noteList[lowerName]
		creature.properName = casterName
	end

	local _, _, difficulty = GetInstanceInfo()
	local spellInfo = creature[difficulty]
	if not spellInfo then
		creature[difficulty] = {}
		spellInfo = creature[difficulty]
	end

	if not spellInfo[spellId] then
		spellInfo[spellId] = true
		if not spellInfo.tooltipString then
			spellInfo.tooltipString = spellName
		else
			-- only add the tooltip text if no other spells are using the same name
			local nameFound
			spellInfo.tooltipString:gsub("([^,]+)[,]?[ ]?", function(name)
				if name == spellName then nameFound = true end
			end)
			if not nameFound then
				spellInfo.tooltipString = spellInfo.tooltipString .. ", " .. spellName
			end
		end
	end
end

----------------------------------------------------------------------------------------------------
-- loading
----------------------------------------------------------------------------------------------------
local function TooltipNotes_OnEventLoad(self, event)
	if TooltipNotesSave           == nil then TooltipNotesSave           = {} end
	if TooltipNotesSave.creatures == nil then TooltipNotesSave.creatures = {} end
	if TooltipNotesSave.settings  == nil then TooltipNotesSave.settings  = {} end

	-- set local references to the creature list and settings
	noteList     = TooltipNotesSave.creatures
	noteSettings = TooltipNotesSave.settings

	-- set default values for the settings if needed
	if noteSettings.wordWrapWidth == nil then noteSettings.wordWrapWidth = 0            end
	if noteSettings.noteColor     == nil then noteSettings.noteColor     = {.8, .8, .8} end
	if noteSettings.spellColor    == nil then noteSettings.spellColor    = {1, 0, 0}    end

	-- start watching combat events
	eventFrame:UnregisterEvent(event)
	eventFrame:SetScript("OnEvent", TooltipNotes_OnEventCombat)
	eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

eventFrame:SetScript("OnEvent", TooltipNotes_OnEventLoad)
eventFrame:RegisterEvent("VARIABLES_LOADED")

----------------------------------------------------------------------------------------------------
-- adding to the tooltip
----------------------------------------------------------------------------------------------------
GameTooltip:HookScript("OnTooltipSetUnit", function(self, tooltip)
	local name, unitid = self:GetUnit()
	if not name or name == "Unknown" or not unitid then
		return
	end

	local creature = UnitExists(unitid) and noteList[name:lower()]
	if creature then
		if not creature.properName then
			creature.properName = name
		end

		-- custom note first
		if creature.customNote then
			local color = noteSettings.noteColor
			if noteSettings.wordWrapWidth ~= 0 then
				local wwNote = WordWrap(creature.customNote, noteSettings.wordWrapWidth)
				for i=1,#wwNote do
					GameTooltip:AddLine(wwNote[i], color[1], color[2], color[3], true)
				end
			else
				GameTooltip:AddLine(creature.customNote, color[1], color[2], color[3], true)
			end
		end

		-- spell list
		if not UnitIsPlayer(unitid) then
			local _, _, difficulty = GetInstanceInfo()
			local spellInfo = creature[difficulty]
			local tooltipString = spellInfo and spellInfo.tooltipString
			if tooltipString then
				local color = noteSettings.spellColor
				if noteSettings.wordWrapWidth ~= 0 then
					local wrappedSpells = WordWrap("Uses: " .. tooltipString, noteSettings.wordWrapWidth)
					for i=1,#wrappedSpells do
						GameTooltip:AddLine(wrappedSpells[i], color[1], color[2], color[3], true)
					end
				else
					GameTooltip:AddLine("Uses: " .. tooltipString, color[1], color[2], color[3], true)
				end
			end
		end
	end
end)
